package ca.uwaterloo.cs.jgrok.env;

import java.util.ArrayList;
import org.xml.sax.Attributes;
import java.lang.reflect.Constructor;

import ca.uwaterloo.cs.jgrok.lib.Function;
import ca.uwaterloo.cs.jgrok.lib.FunctionLib;
import ca.uwaterloo.cs.jgrok.lib.FunctionLibManager;

public class FunctionLoader extends DefaultLoader {
    private String fName;
    private String fType;
    private String fClass;

    private int argCount = 0;
    private ArrayList<Argument> args;
    
    private int paramCount = 0;
    private ArrayList<Param> params;
    
    private FunctionLib fLib = null;
    
    public FunctionLoader() {
        fLib = FunctionLib.rootLib();
        args = new ArrayList<Argument>();
        params = new ArrayList<Param>();
    }
    
    public void startElement(String uri, String name, String qName, Attributes atts) {
        String value;
        String localName;
        
        // Process FunctionLib
        if (name.equalsIgnoreCase("FunctionLib")) {
            String libName;
            
            libName = atts.getValue(uri, "Name");
            if(libName == null) {
                libName = FunctionLibManager.JGrokLibName;
            }
            
            fLib = env.getFunctionLibManager().getLib(libName);
        }
        // Process Function
        else if (name.equalsIgnoreCase("Function")) {
            fName = null;
            fType = null;
            fClass = null;
            args.clear();
            argCount = 0;
            params.clear();
            paramCount = 0;
            
            for(int i = 0; i < atts.getLength(); i++) {
                value = atts.getValue(i);
                localName = atts.getLocalName(i);
                
                if(localName.equalsIgnoreCase("Name"))
                    fName = value;
                else if(localName.equalsIgnoreCase("Type"))
                    fType = value;
                else if(localName.equalsIgnoreCase("Class"))
                    fClass = value;
                else
                    System.out.println("Start element: {" + uri + "}" + name);
            }
        }
        // Process Constructor
        else if (name.equalsIgnoreCase("Constructor")) {
            args.clear();
            argCount = 0;
        }
        // Process Arguments for constructing Function
        else if (name.equalsIgnoreCase("Arguments")) {
            value = atts.getValue(uri, "Count");
            argCount = Integer.parseInt(value);
        }
        // Process Argument
        else if (name.equalsIgnoreCase("Argument")) {
            Argument a = new Argument();
            
            for(int i = 0; i < atts.getLength(); i++) {
                value = atts.getValue(i);
                localName = atts.getLocalName(i);
                
                if(localName.equalsIgnoreCase("Name"))
                    a.name = value;
                else if(localName.equalsIgnoreCase("Type"))
                    a.type = value;
                else if(localName.equalsIgnoreCase("Value"))
                    a.value = value;
                else
                    System.out.println("Start element: {" + uri + "}" + name);
            }
            args.add(a);
        }
        // Process Parameters
        else if (name.equalsIgnoreCase("Parameters")) {
            value = atts.getValue(uri, "Count");
            paramCount = Integer.parseInt(value);
        }
        // Process Parameter
        else if (name.equalsIgnoreCase("Parameter")) {
            Param p = new Param();
            
            for(int i = 0; i < atts.getLength(); i++) {
                value = atts.getValue(i);
                localName = atts.getLocalName(i);
                
                if(localName.equalsIgnoreCase("Name"))
                    p.name = value;
                else if(localName.equalsIgnoreCase("Type"))
                    p.type = value;
                else
                    System.out.println("Start element: {" + uri + "}" + name);
            }
            params.add(p);
        }
    }

    public void endElement(String uri, String name, String qName) {
                
        if (name.equalsIgnoreCase("Function")) {
            if (argCount != args.size()) {
                return;
            }
            if (paramCount != params.size()) {
                return;
            }
            
            try {
                Function func = null;
                Class<? extends Function> defClass;
                Class<?>[] paramTypes = new Class[paramCount];
                
                defClass = Class.forName(fClass).asSubclass(Function.class);
                
                if (argCount == 0) {
                    func = defClass.newInstance();
                } else {
                    Constructor<? extends Function> c;
                    Object[] valueArray = new Object[args.size()];
                    Class<?>[] typeArray = new Class<?>[args.size()];
                    
                    ArrayList<Class<?>> typeList = new ArrayList<Class<?>>(args.size());
                    for(Argument a : args) {
                        typeList.add(a.getType());
                    }
                    
                    ArrayList<Object> valueList = new ArrayList<Object>(args.size());
                    for(Argument a : args) {
                        valueList.add(a.getValue());
                    }
                    
                    typeList.toArray(typeArray);
                    c = defClass.getConstructor(typeArray);
                    
                    valueList.toArray(valueArray);
                    func = c.newInstance(valueArray);
                }
                
                // Set function name
                func.setName(fName);
                
                // Set function return type
                if(fType != null) {
                    func.setType(Clazz.forName(fType));
                }
                
                // Set function parameter types
                if(paramCount > 0) {
                    for(int i = 0; i < params.size(); i++) {
                        paramTypes[i] = params.get(i).getType();
                    }
                    func.setParamTypes(paramTypes);
                }
                
                fLib.register(func);
                
                Tracing trc = env.getTracing();
                if (trc.traceOn(TracingEvent.TRACING_FUNCTIONLIB)) {
                    trc.out.println("\t" + func);
                }
            } catch(Exception e) {
               e.printStackTrace();
               System.err.println(e + "  message=" +e.getMessage());
            }
        }
    }
    
    static class Param {
        String name;
        String type;
        
        Param() {
            this.name = null;
            this.type = null;
        }
        
        Param(String name, String type) {
            this.name = name;
            this.type = type;
        }
        
        String getName() {
            return name;
        }
        
        Class<?> getType() throws ClassNotFoundException {
            return Clazz.forName(type);
        }
    }
    
    static class Argument {
        String name;
        String type;
        String value;
        
        Argument() {
            this.name = null;
            this.type = null;
            this.value = null;
        }
        
        Argument(String name, String type, String value) {
            this.name = name;
            this.type = type;
        }
        
        String getName() {
            return name;
        }
        
        Class<?> getType() throws ClassNotFoundException {
            return Clazz.forName(type);
        }
        
        Object getValue() throws ClassNotFoundException, InstantiationException {
            return Clazz.forValue(getType(), value);
        }
    }
}
