package ca.uwaterloo.cs.jgrok.interp;

import java.io.PrintStream;
import ca.uwaterloo.cs.jgrok.env.Clazz;
import ca.uwaterloo.cs.jgrok.fb.TupleSet;

public class Value {
    Object val;
    Class<?> primitiveType;
    
    public static final Value VOID = new Value();
    public static final Value EVAL = new Value();
    
    static {
        VOID.primitiveType = void.class;
        EVAL.primitiveType = void.class;
    }
    
    protected Value() {
        val = new Object();
        primitiveType = null;
    }
    
    public Value(int i) {
        val = new Integer(i);
        primitiveType = int.class;
    }
    
    public Value(long l) {
        val = new Long(l);
        primitiveType = long.class;
    }
    
    public Value(short s) {
        val = new Short(s);
        primitiveType = short.class;
    }
    
    public Value(float f) {
        val = new Float(f);
        primitiveType = float.class;
    }

    public Value(double d) {
        val = new Double(d);
        primitiveType = double.class;
    }
    
    public Value(boolean b) {
        val = new Boolean(b);
        primitiveType = boolean.class;
    }
    
    public Value(Object o) {
        val = o;
        primitiveType = null;
    }
    
    public Value(Object o, Class<?> clazz) {
        val = o;
        if(Clazz.isJavaPrimitive(clazz)) {
            primitiveType = clazz;
        } else {
            primitiveType = null;
        }
    }
    
    public Class<?> getType() {
        if(primitiveType != null) {
            return primitiveType;
        } else {
            return val.getClass();
        }
    }
    
    /**
     * Tests Primitives and String.
     */
    public boolean isPrimitive() {
        if(primitiveType != null) return true;
        else if(val.getClass() == String.class) return true;
        else return false;
    }
    
    public String toString() {
        if(val == null) return "null";
        else return val.toString();
    }
    
    public int intValue() {
        if(primitiveType == short.class) {
            return ((Short)val).shortValue();
        } else {
            return ((Integer)val).intValue();
        }
    }
    
    public long longValue() {
        if(primitiveType == short.class) {
            return ((Short)val).shortValue();
        } else if(primitiveType == int.class){
            return ((Integer)val).intValue();
        } else {
            return ((Long)val).longValue();
        }
    }
    
    public short shortValue() {
        return ((Short)val).shortValue();
    }
    
    public float floatValue() {
        if(primitiveType == int.class) {
            return ((Integer)val).intValue();
        } else if(primitiveType == long.class) {
            return ((Long)val).longValue();
        } else if(primitiveType == short.class) {
            return ((Short)val).shortValue();
        } else {
            return ((Float)val).floatValue();
        }
    }
    
    public double doubleValue() {
        if(primitiveType == int.class) {
            return ((Integer)val).intValue();
        } else if(primitiveType == long.class) {
            return ((Long)val).longValue();
        } else if(primitiveType == short.class) {
            return ((Short)val).shortValue();
        } else if(primitiveType == float.class) {
            return ((Float)val).floatValue();
        } else {
            return ((Double)val).doubleValue();
        }
    }
    
    public boolean booleanValue() {
        return ((Boolean)val).booleanValue();
    }
    
    public Object objectValue() {
        return val;
    }
    
    public void set(int i) {
        val = new Integer(i);
        primitiveType = int.class;
    }

    public void set(long l) {
        val = new Long(l);
        primitiveType = long.class;
    }
    
    public void set(short s) {
        val = new Short(s);
        primitiveType = short.class;
    }
    
    public void set(float f) {
        val = new Float(f);
        primitiveType = float.class;
    }

    public void set(double d) {
        val = new Double(d);
        primitiveType = double.class;
    }
    
    public void set(boolean b) {
        val = new Boolean(b);
        primitiveType = boolean.class;
    }
    
    public void set(Object o) {
        val = o;
        primitiveType = null;
    }
    
    public void print(PrintStream ps) {
        if(isPrimitive()) {
            if(primitiveType != void.class) {
                ps.println(this);
            }
        } else if(val instanceof Value[]) {
            Value[] vals = (Value[])val;
            for(int i = 0; i < vals.length; i++) {
                vals[i].print(ps);
            }
        } else {
            if(Type.isSubtypeOf(getType(), TupleSet.class)) {
                ((TupleSet)val).print(ps);
            } else {
                ps.println(val);
            }
        }
    }
}
