package ca.uwaterloo.cs.jgrok.lib;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.Value;

/**
 * <pre>
 *     EdgeSet degree(TupleSet, col)
 *     EdgeSet indegree(TupleSet, col)
 *     EdgeSet outdegree(TupleSet, col)
 *
 *     TupleSet degree(TupleSet, col, ...)
 * </pre>
 */
public class Degree extends Function {
    
    public Degree(String functionName) {
        if(functionName == null) {
            this.name = "degree";
            type = EdgeSet.class;
            paramTypes = new Class[2];
            paramTypes[0] = TupleSet.class;
            paramTypes[1] = int.class;
        } else {
            this.name = functionName;
        }
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        int column;
        TupleSet tSet;
        EdgeSet result;
        
        tSet = (TupleSet)vals[0].objectValue();
        column = vals[1].intValue();
        
        if(name.equals("degree")) {
            if(vals.length == 2) {
                result = UtilityOperation.degree(tSet, column);
            } else if(vals.length > 2) {
                int[] cols = new int[vals.length-1];
                for(int i = 0; i < vals.length-1; i++) {
                    cols[i] = vals[i+1].intValue();
                }
                return new Value(UtilityOperation.degree(tSet, cols));
            } else {
                return new Value(new TupleSet());
            }
        } else if (name.equals("indegree")) {
            result = UtilityOperation.indegree(tSet, column);
        } else {
            result = UtilityOperation.outdegree(tSet, column);
        }
        
        return new Value(result);
    }
}
