package ca.uwaterloo.cs.jgrok.lib;

import java.io.*;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.interp.Value;

/**
 * void exe(...)
 * void run(...)
 */
public class Execute extends Function {
    
    public Execute() {
        name = "exe";
        type = void.class;
        paramTypes = new Class<?>[1];
        paramTypes[0] = null;
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        if(vals.length == 0) {
            env.out.println("[Warning] Execute nothing");
            return Value.VOID;
        }
        
        String[] cmdArray = new String[vals.length];
        for(int i = 0; i < cmdArray.length; i++) {
            cmdArray[i] = vals[i].objectValue().toString();
        }
        
        if(name.equals("run")) {
            ExecRunnable run = new ExecRunnable(cmdArray);
            new Thread(run).start();
        } else {
            wait(env, cmdArray);
        }
        return Value.VOID;
    }

    private void wait(Env env, String[] cmdArray) {
        Process p;
        
        try {
            if(cmdArray.length == 1) {
                p = Runtime.getRuntime().exec(cmdArray[0]);
            } else {
                p = Runtime.getRuntime().exec(cmdArray);
            }
            
            String line;
            BufferedReader reader;
            reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
            while((line = reader.readLine()) != null) {
                env.out.println(line);
            }
            
            reader = new BufferedReader(new InputStreamReader(p.getErrorStream()));
            while((line = reader.readLine()) != null) {
                env.err.println(line);
            }
        } catch(Exception e) {
            e.printStackTrace(env.err);
        }
    }
}

class ExecRunnable implements Runnable {
    private String[] cmdArray;

    ExecRunnable(String[] cmdArray) {
        this.cmdArray = cmdArray;
    }
    
    public void run() {
        Process p;
        
        try {
            if(cmdArray.length == 1) {
                p = Runtime.getRuntime().exec(cmdArray[0]);
            } else {
                p = Runtime.getRuntime().exec(cmdArray);
            }
            
            String line;
            BufferedReader reader;
            reader = new BufferedReader(new InputStreamReader(p.getInputStream()));
            while((line = reader.readLine()) != null) {
                System.out.println(line);
            }
            
            reader = new BufferedReader(new InputStreamReader(p.getErrorStream()));
            while((line = reader.readLine()) != null) {
                System.err.println(line);
            }
        } catch(Exception e) {
            e.printStackTrace(System.err);
        }
    }
}
