package ca.uwaterloo.cs.jgrok.lib;

import java.util.HashMap;

public class FunctionLibManager {
    public static final String JGrokLibName = "JGrok";
    public static final FunctionLib JGrokLib = new FunctionLib(JGrokLibName);
    
    private FunctionLib rootLib;
    private HashMap<String,FunctionLib> libEntries;
    
    /**
     * Creates a FunctionLib manager.
     */
    public FunctionLibManager() {
        rootLib = JGrokLib;
        libEntries = new HashMap<String,FunctionLib>();
        libEntries.put(rootLib.getName(), rootLib);
    }
    
    /**
     * Gets the root FunctionLib.
     */
    public FunctionLib getRootLib() {
        return rootLib;
    }
    
    /**
     * Finds a FunctionLib by name.
     * @return the found FunctionLib or null if no FunctionLib was found.
     */
    public FunctionLib findLib(String libName) {
        return libEntries.get(libName);
    }
    
    /**
     * Gets a FunctionLib by name.
     * If no FunctionLib was found, a new one is created.
     */
    public FunctionLib getLib(String libName) {
        FunctionLib lib = findLib(libName);
        if(lib == null) {
            lib = new FunctionLib(libName);
            libEntries.put(libName, lib);
        }
        
        return lib;
    }
    
    /**
     * Registers a function.
     * 
     * @param libName the function library
     * @param f the function to be registered
     * @return <b>true</b> if successfully registered.
     */
    public boolean register(String libName, Function f) {
        FunctionLib lib = getLib(libName);
        return lib.register(f);
    }
}
