package ca.uwaterloo.cs.jgrok.lib;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.Value;

/**
 * <pre>
 *     EdgeSet indegree(EdgeSet)
 *     EdgeSet indegree(NodeSet, EdgeSet)
 *
 *     EdgeSet outdegree(EdgeSet)
 *     EdgeSet outdegree(NodeSet, EdgeSet)
 * </pre>
 */
public class InDegree extends Function {
    
    public InDegree(String functionName) {
        if (functionName == null) {
            this.name = "indegree";
        } else {
            this.name = functionName;
        }
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        NodeSet nSet;
        EdgeSet eSet;
        EdgeSet result;
        
        if(vals.length == 1) {
            eSet = (EdgeSet)vals[0].objectValue();
        } else {
            eSet = (EdgeSet)vals[1].objectValue();
        }
        
        if(name.equals("indegree")) {
            if(vals.length == 1) {
                nSet = AlgebraOperation.rangeOf(eSet);
            } else {
                nSet = (NodeSet)vals[0].objectValue();
            }
            result = AlgebraOperation.composition(eSet, nSet);
            result = AlgebraOperation.inverse(result);
            result = UtilityOperation.outdegree(result);
        } else {
            if(vals.length == 1) {
                nSet = AlgebraOperation.domainOf(eSet);
            } else {
                nSet = (NodeSet)vals[0].objectValue();
            }
            
            result = AlgebraOperation.composition(nSet, eSet);
            result = UtilityOperation.outdegree(result);
        }
        
        return new Value(result);
    }
}
