package ca.uwaterloo.cs.jgrok.lib;

import java.util.Arrays;
import java.util.ArrayList;
import java.util.Enumeration;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;

/**
 * <pre>
 * Functions:
 *
 *     NodeSet list()
 *     NodeSet flist()
 *     NodeSet relnames()
 *     NodeSet setnames()
 *
 * </pre> 
 */
public class Listing extends Function {
    
    public Listing() {
        name = "list";
        type = NodeSet.class;
        paramTypes = new Class[0];
    }
    
    public Value invoke(Env env, Value[] vals) throws InvocationException {
        if(name.equals("list")) {
            return action_list(env);
        } else if(name.equals("flist")) {
            return action_flist(env);
        } else if(name.equals("relnames")) {
            return action_relnames(env);
        } else if(name.equals("setnames")) {
            return action_setnames(env);
        } else {
            throw InvocationException.createUnknownFunction(getName());
        }
    }
    
    private Value action_list(Env env) {
        NodeSet set = new NodeSet();        
        Scope scp = env.peepScope();
        Enumeration<Variable> enm = scp.allVariables();
        ArrayList<String> list = new ArrayList<String>();
        
        Variable var;
        while(enm.hasMoreElements()) {
            var = enm.nextElement();
            list.add(var.getName());
        }
        
        // Sort names
        String[] names = new String[list.size()];
        list.toArray(names);
        Arrays.sort(names);
        
        // Save names to set
        for(int i = 0; i < names.length; i++) {
            set.add(names[i]);
        }
        
        return new Value(set);
    }
    
    private Value action_flist(Env env) {
        String[] names;
        FunctionLib fLib;
        NodeSet set = new NodeSet();
        
        fLib = FunctionLib.rootLib();
        names = fLib.getFunctionNames();
        Arrays.sort(names);
        
        for(int i = 0; i < names.length; i++) {
            set.add(names[i]);
        }
        
        return new Value(set);
    }
    
    private Value action_relnames(Env env) {
        NodeSet set = new NodeSet();
        Scope scp = env.peepScope();
        Enumeration<Variable> enm = scp.allVariables();
        ArrayList<String> list = new ArrayList<String>();
        
        Class<?> type;
        Variable var;
        while(enm.hasMoreElements()) {
            var = enm.nextElement();
            type = var.getType();
            if(type == EdgeSet.class ||
               type == TupleSet.class ) {
                list.add(var.getName());
            }
        }
        
        // Sort names
        String[] names = new String[list.size()];
        list.toArray(names);
        Arrays.sort(names);
        
        // Save names to set
        for(int i = 0; i < names.length; i++) {
            set.add(names[i]);
        }
        
        return new Value(set);
    }

    private Value action_setnames(Env env) {
        NodeSet set = new NodeSet();
        Scope scp = env.peepScope();
        Enumeration<Variable> enm = scp.allVariables();
        ArrayList<String> list = new ArrayList<String>();
        
        Class<?> type;
        Variable var;
        while(enm.hasMoreElements()) {
            var = enm.nextElement();
            type = var.getType();
            if(type == NodeSet.class) {
                list.add(var.getName());
            }
        }
        
        // Sort names
        String[] names = new String[list.size()];
        list.toArray(names);
        Arrays.sort(names);
        
        // Save names to set
        for(int i = 0; i < names.length; i++) {
            set.add(names[i]);
        }
        
        return new Value(set);
    }
}
