package ca.uwaterloo.cs.jgrok.lib;

import java.util.Enumeration;
import java.io.FileNotFoundException;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;
import ca.uwaterloo.cs.jgrok.io.*;
import ca.uwaterloo.cs.jgrok.io.ta.TAFileReader;

/**
 * <pre>
 *    void read (string dataFile)
 *    void getta(string dataFile)
 *    void getdb(string dataFile)
 * </pre>
 */
public class Read extends Function {
    
    public Read(String functionName) {
        name = functionName;
        type = void.class;
        paramTypes = new Class[1];
        paramTypes[0] = String.class;
    }
    
    public Value invoke(Env env, Value[] vals) throws InvocationException {
        String fileName;
        FactbaseReader fbReader;
        fileName = vals[0].toString();
        
        if(name.equals("read")) {
            fbReader = getReader(fileName);
        } else if(name.equals("getta")) {
            fbReader = new TAFileReader();
        } else if(name.equals("getdb")) {
            fbReader = new RSFFileReader();
        } else {
            throw new InvocationException("illegal use: " + name);
        }
        
        return load(env, fbReader, fileName);
    }
    
    private FactbaseReader getReader(String fileName)
        throws InvocationException {
        FactbaseReader fbReader;
        
        if(fileName.endsWith(".ta")) {
            fbReader = new TAFileReader();
        } else if(fileName.endsWith(".rsf")) {
            fbReader = new RSFFileReader();
        } else {
            throw new InvocationException("illegal file: " + fileName);
        }
        
        return fbReader;
    }
    
    private Value load(Env env, FactbaseReader fbReader, String fileName)
        throws InvocationException {
        Factbase fb;
        
        try {
            fb = fbReader.read(fileName);
        } catch (FileNotFoundException e) {
            throw new InvocationException(fileName + " not found");
        }
        
        Scope scp;
        Variable var;
        TupleSet set;
        Enumeration<TupleSet> enm;
        
        EdgeSet eSet1;
        EdgeSet eSet2;
        EdgeSet result;
        
        enm = fb.allSets();
        scp = env.peepScope();
        while(enm.hasMoreElements()) {
            set = enm.nextElement();
            try {
                var = scp.lookup(set.getName());
                if(set instanceof EdgeSet &&
                   var.getType() == EdgeSet.class) {
                    eSet1 = (EdgeSet)var.getValue().objectValue();
                    eSet2 = (EdgeSet)set;
                    result= AlgebraOperation.union(eSet1, eSet2);
                    var.setValue(new Value(result));
                    result.setName(var.getName());
                } else {
                    var.setValue(new Value(set));
                }
            } catch(LookupException e) {
                var = new Variable(scp, set.getName(), new Value(set));
                scp.addVariable(var);
            }
        }
            
        return Value.VOID;
    }
}
