package ca.uwaterloo.cs.jgrok.lib;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;
import ca.uwaterloo.cs.jgrok.util.*;

/**
 * SimRank for bipartite graph.
 * <pre>
 *     TupleSet simBi(EdgeSet biGraph)
 *     TupleSet simBi(EdgeSet biGraph, int iterationCount)
 *
 *     TupleSet simBi(EdgeSet biGraph, EdgeSet featureGraph)
 *     TupleSet simBi(EdgeSet biGraph, EdgeSet featureGraph, int iterationCount)
 * </pre>
 */
public class SimRankBi extends Function {
    
    public SimRankBi() {
        name = "simBi";
        type = TupleSet.class;
        paramTypes = new Class[1];
        paramTypes[0] = EdgeSet.class;
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        EdgeSet graph = (EdgeSet)vals[0].objectValue();
        SimRankBipartite simRank = new SimRankBipartite(graph);
        SimInitialization simInit = null;
        
        int iterationCount = 10;        
        
        if(vals.length == 2) {
            Object o = vals[1].objectValue();
            if(o instanceof EdgeSet)
                simInit = new SimInitialization((EdgeSet)o, null); 
            else iterationCount = vals[1].intValue();
        } else if(vals.length == 3) {
            simInit = new SimInitialization((EdgeSet)vals[1].objectValue(), null);
            iterationCount = vals[2].intValue();
        }
        
        if(simInit != null) simInit.initialize(simRank);
        
        return new Value(simRank.compute(iterationCount));
    }
}
