package ca.uwaterloo.cs.jgrok.lib.math;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;
import ca.uwaterloo.cs.jgrok.lib.Function;
import ca.uwaterloo.cs.jgrok.lib.InvocationException;

/**
 * <pre>
 * Functions:
 *
 *     float avg(NodeSet)
 *     float avg(TupleSet, col)
 *
 * Examples:
 *     A { age = 10 }
 *     B { age = 10 }
 *     C { age = 30 }
 *     D { age = 32 }
 *     E { age = 32 }
 *
 *     >> avg(rng @age)
 *     24.0
 *     >> avg(@age, 1)
 *     22.8
 *
 * </pre>
 */
public class Avg extends Function {
    
    public Avg() {
        name = "avg";
        type = float.class;
        paramTypes = new Class[1];
        paramTypes[0] = NodeSet.class;
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        float result;
        TupleSet input;
        
        result = 0;
        input = (TupleSet)vals[0].objectValue();
        
        if(input.size() > 0) {
            TupleList data = input.getTupleList();
            
            if(vals.length == 1) {
                for(int i = 0; i < data.size(); i++) {
                    Tuple tuple = data.get(i);
                    try {
                        result += Float.parseFloat(IDManager.get(tuple.getDom()));
                    } catch(NumberFormatException e) {}
                }
            } else {
                int col = vals[1].intValue();
                for(int i = 0; i < data.size(); i++) {
                    Tuple tuple = data.get(i);
                    if(col < tuple.size()) {
                        try {
                            result += Float.parseFloat(IDManager.get(tuple.get(col)));
                        } catch(NumberFormatException e) {}
                    } else {
                        throw new InvocationException("index " + col + " out of bounds " + tuple.size());
                    }
                }
            }
            
            result = result / input.size();
        }
        
        return new Value(result);
    }
}
